using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Net;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Class1;

namespace WebDAVServer.Class1Listener
{
    /// <summary>
    /// Represents file in WebDAV repository.
    /// </summary>
    public class DavFile : DavHierarchyItem, IFile
    {

        /// <summary>
        /// Initializes a new instance of the DavFile class.
        /// </summary>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        /// <param name="itemId">Item identifier.</param>
        /// <param name="parentId">Parent identifier.</param>
        /// <param name="name">Item name.</param>
        /// <param name="path">Item path (encoded)</param>
        /// <param name="created">Item creation date.</param>
        /// <param name="modified">Item modification date.</param>
        /// <param name="fileAttributes">File attributes for the item.</param>
        public DavFile(
            DavContext context,
            Guid itemId,
            Guid parentId,
            string name,
            string path,
            DateTime created,
            DateTime modified)
            : base(context, itemId, parentId, name, path, created, modified)
        {
        }

        /// <summary>
        /// Gets item's content type.
        /// </summary>
        public string ContentType
        {
            get
            {
                string contentType = getDbField<string>("ContentType", null);

                if (string.IsNullOrEmpty(contentType))
                {
                    contentType = MimeType.GetMimeType(System.IO.Path.GetExtension(Name)) ?? "application/octet-stream";
                }

                return contentType;
            }
        }

        /// <summary>
        /// Gets file length.
        /// </summary>
        public long ContentLength
        {
            get
            {
                object result = Context.ExecuteScalar<object>(
                    "SELECT DATALENGTH(Content) FROM Item WHERE ItemId = @ItemId",
                    "@ItemId", ItemId);

                return result == DBNull.Value ? 0 : Convert.ToInt64(result);
            }
        }

        /// <summary>
        /// Gets Etag.
        /// </summary>
        public string Etag
        {
            get
            {
                int serialNumber = getDbField("SerialNumber", 0);
                return string.Format("{0}-{1}", Modified.ToBinary(), serialNumber);
            }
        }

        /// <summary>
        /// Reads file's content from storage (to send to client).
        /// </summary>
        /// <param name="output">Stream to read body to.</param>
        /// <param name="byteStart">Number of first byte to write.</param>
        /// <param name="count">Number of bytes to be written.</param>
        public void Read(Stream output, long byteStart, long count)
        {

            if (Context.Request.RawUrl.EndsWith("?download"))
            {
                AddContentDisposition();
            }

            using (SqlDataReader reader = Context.ExecuteReader(
                CommandBehavior.SequentialAccess,
                @"SELECT Content FROM Item WHERE ItemId = @ItemId",
                "@ItemId", ItemId))
            {
                reader.Read();

                long bufSize = 1048576; // 1Mb
                var buf = new byte[bufSize];
                long retval;
                long startIndex = byteStart;

                while ((retval = reader.GetBytes(
                    reader.GetOrdinal("Content"),
                    startIndex,
                    buf,
                    0,
                    (int)(count > bufSize ? bufSize : count))) > 0)
                {
                    output.Write(buf, 0, (int)retval);
                    startIndex += retval;
                    count -= retval;
                }
            }
        }

        private void AddContentDisposition()
        {
            if (Context.Request.UserAgent.Contains("MSIE")) // problem with file extensions in IE
            {
                var fileName = EncodeUtil.EncodeUrlPart(Name);
                Context.Response.AddHeader("Content-Disposition", "attachment; filename=\"" + fileName + "\"");
            }
            else
                Context.Response.AddHeader("Content-Disposition", "attachment;");
        }

        /// <summary>
        /// Stores file contents to storage (when client updates it).
        /// </summary>
        /// <param name="segment">Stream with new file content.</param>
        /// <param name="contentType">New content type.</param>
        /// <param name="startIndex">Index of first byte in the file where update shall be applied.</param>
        /// <param name="totalContentLength">Length of the file after it will be updated with the new content.</param>
        /// <returns>Boolean value indicating if entire stream was written.</returns>
        public bool Write(Stream segment, string contentType, long startIndex, long totalContentLength)
        {

            string commandText = 
                @"UPDATE Item
                  SET
                      Modified = GETUTCDATE(),
                      ContentType = @ContentType,
                      Content = CASE WHEN @ResetContent = 1 THEN 0x ELSE Content END,
                      SerialNumber = ISNULL(SerialNumber, 0) + 1
                  WHERE ItemId = @ItemId";

            Context.ExecuteNonQuery(
                 commandText,
                 "@ContentType", (object)contentType ?? DBNull.Value,
                 "@ItemId", ItemId,
                 "@ResetContent", startIndex == 0 ? 1 : 0);

            const int bufSize = 1048576; // 1Mb
            byte[] buffer = new byte[bufSize];

            long bytes = 0;

            int lastBytesRead;

            while ((lastBytesRead = segment.Read(buffer, 0, bufSize)) > 0)
            {
                SqlParameter dataParm = new SqlParameter("@Data", SqlDbType.VarBinary, bufSize);
                SqlParameter offsetParm = new SqlParameter("@Offset", SqlDbType.Int);
                SqlParameter bytesParm = new SqlParameter("@Bytes", SqlDbType.Int, bufSize);
                SqlParameter itemIdParm = new SqlParameter("@ItemId", ItemId);

                dataParm.Value = buffer;
                dataParm.Size = lastBytesRead;
                offsetParm.Value = bytes + startIndex;
                bytesParm.Value = lastBytesRead;
                
                string updateItemCommand =
                        @"UPDATE Item SET
                            Content .WRITE(@Data, @Offset, @Bytes)
                          WHERE ItemId = @ItemId";

                Context.ExecuteNonQuery(
                        updateItemCommand,
                        dataParm,
                        offsetParm,
                        bytesParm,
                        itemIdParm);

                bytes += lastBytesRead;
            }

            return true;
        }

        /// <summary>
        /// Copies this file to another folder.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">New file name in destination folder.</param>
        /// <param name="deep">Is not used.</param>
        /// <param name="multistatus">Container for errors with items other than this file.</param>
        public override void CopyTo(
            IItemCollection destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus)
        {
            DavFolder destDavFolder = destFolder as DavFolder;
            if (destFolder == null)
            {
                throw new DavException("Destination folder doesn't exist", DavStatus.CONFLICT);
            }

            DavHierarchyItem destItem = destDavFolder.FindChild(destName);
            if (destItem != null)
            {
                try
                {
                    destItem.Delete(multistatus);
                }
                catch (DavException ex)
                {
                    multistatus.AddInnerException(destItem.Path, ex);
                    return;
                }
            }

            CopyThisItem(destDavFolder, null, destName);
        }

        /// <summary>
        /// Deletes this file.
        /// </summary>
        /// <param name="multistatus">Is not used.</param>
        public override void Delete(MultistatusException multistatus)
        {
            DavFolder parent = GetParent();
            if (parent == null)
            {
                throw new DavException("Parent is null", DavStatus.CONFLICT);
            }

            DeleteThisItem(parent);
        }

        /// <summary>
        /// Moves this file to different folder and renames it.
        /// </summary>
        /// <param name="destFolder">Destination folder.</param>
        /// <param name="destName">New file name.</param>
        /// <param name="multistatus">Container for errors with items other than this file.</param>
        public override void MoveTo(IItemCollection destFolder, string destName, MultistatusException multistatus)
        {
            DavFolder destDavFolder = destFolder as DavFolder;
            if (destFolder == null)
            {
                throw new DavException("Destination folder doesn't exist", DavStatus.CONFLICT);
            }

            DavFolder parent = GetParent();
            if (parent == null)
            {
                throw new DavException("Cannot move root.", DavStatus.CONFLICT);
            }

            DavHierarchyItem destItem = destDavFolder.FindChild(destName);
            if (destItem != null)
            {
                try
                {
                    destItem.Delete(multistatus);
                }
                catch (DavException ex)
                {
                    multistatus.AddInnerException(destItem.Path, ex);
                    return;
                }
            }

            MoveThisItem(destDavFolder, destName, parent);
        }

        /// <summary>
        /// Returns database field of the file.
        /// </summary>
        /// <typeparam name="T">Type of field value.</typeparam>
        /// <param name="columnName">DB columen in which field is stored.</param>
        /// <param name="defaultValue">Default value to return.</param>
        /// <returns>Value from database or default value if it is null.</returns>
        private T getDbField<T>(string columnName, T defaultValue)
        {
            string commandText = string.Format("SELECT {0} FROM Item WHERE ItemId = @ItemId", columnName);
            object obj = Context.ExecuteScalar<object>(commandText, "@ItemId", ItemId);

            return obj != null ? (T)obj : defaultValue;
        }
    }
}
