using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Class1;

namespace WebDAVServer.Class1Listener
{
    /// <summary>
    /// Base class for items like files, folders, versions etc.
    /// </summary>
    public abstract class DavHierarchyItem : IHierarchyItem
    {
        protected DavContext Context { get; private set; }

        public Guid ItemId { get; private set; }

        public string Name { get; private set; }

        public string Path { get; private set; }

        public DateTime Created { get; private set; }

        public DateTime Modified { get; private set; }

        protected Guid ParentId { get; private set; }

        public DavHierarchyItem(
            DavContext context,
            Guid itemId,
            Guid parentId,
            string name,
            string path,
            DateTime created,
            DateTime modified)
        {
            Context = context;
            ItemId = itemId;
            ParentId = parentId;
            Name = name;
            Path = path;
            Created = created;
            Modified = modified;
        }

        public DavFolder GetParent()
        {
            string[] parts = Path.TrimEnd('/').Split('/');
            string parentParentPath = "/";
            if (parts.Length > 2)
            {
                parentParentPath = string.Join("/", parts, 0, parts.Length - 2);
            }

            string command =
                @"SELECT 
                     ItemID
                   , ParentItemId
                   , ItemType
                   , Name
                   , Created
                   , Modified
                                     FROM Item
                  WHERE ItemId = @ItemId";

            return Context.ExecuteItem<DavFolder>(
                parentParentPath,
                command,
                "@ItemId", ParentId).FirstOrDefault();
        }

        public abstract void CopyTo(
            IItemCollection destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus);

        public abstract void MoveTo(IItemCollection destFolder, string destName, MultistatusException multistatus);

        public abstract void Delete(MultistatusException multistatus);

        public IEnumerable<PropertyValue> GetProperties(IList<PropertyName> names, bool allprop)
        {
            IList<PropertyValue> propVals =
                Context.ExecutePropertyValue(
                    "SELECT Name, Namespace, PropVal FROM Property WHERE ItemID = @ItemID",
                    "@ItemID", ItemId);

            foreach (PropertyValue p in propVals)
            {
                if (allprop || names.Contains(p.QualifiedName))
                {
                    yield return p;
                }
            }
        }
        
        public virtual void UpdateProperties(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus)
        {

            foreach (PropertyValue p in setProps)
            {
               SetProperty(p); // create or update property
            }

            foreach (PropertyName p in delProps)
            {
               RemoveProperty(p.Name, p.Namespace);
            }

            // You should not update modification date/time here. Mac OS X Finder expects that properties update do not change the file modification date.
        }

        public IEnumerable<PropertyName> GetPropertyNames()
        {
            foreach (PropertyValue val in GetProperties(new PropertyName[0], true))
            {
                yield return val.QualifiedName;
            }            
        }
        
        protected void SetProperty(PropertyValue prop)
        {
            string selectCommand = 
                @"SELECT Count(*) FROM Property
                  WHERE ItemID = @ItemID AND Name = @Name AND Namespace = @Namespace";

            int count = Context.ExecuteScalar<int>(
                selectCommand,
                "@ItemID", ItemId,
                "@Name", prop.QualifiedName.Name,
                "@Namespace", prop.QualifiedName.Namespace);

            // insert
            if (count == 0)
            {
                string insertCommand = @"INSERT INTO Property(ItemID, Name, Namespace, PropVal)
                                          VALUES(@ItemID, @Name, @Namespace, @PropVal)";

                Context.ExecuteNonQuery(
                    insertCommand,
                    "@PropVal", prop.Value,
                    "@ItemID", ItemId,
                    "@Name", prop.QualifiedName.Name,
                    "@Namespace", prop.QualifiedName.Namespace);
            }
            else
            {
                // update
                string command = @"UPDATE Property
                      SET PropVal = @PropVal
                      WHERE ItemID = @ItemID AND Name = @Name AND Namespace = @Namespace";

                Context.ExecuteNonQuery(
                    command,
                    "@PropVal", prop.Value,
                    "@ItemID", ItemId,
                    "@Name", prop.QualifiedName.Name,
                    "@Namespace", prop.QualifiedName.Namespace);
            }
        }

        protected void RemoveProperty(string name, string ns)
        {
            string command = @"DELETE FROM Property
                              WHERE ItemID = @ItemID
                              AND Name = @Name
                              AND Namespace = @Namespace";

            Context.ExecuteNonQuery(
                command,
                "@ItemID", ItemId,
                "@Name", name,
                "@Namespace", ns);
        }

        internal DavFolder CopyThisItem(DavFolder destFolder, DavHierarchyItem destItem, string destName)
        {
            // returns created folder, if any, otherwise null
            DavFolder createdFolder = null;

            Guid destID;
            if (destItem == null)
            {
                // copy item
                string commandText =
                    @"INSERT INTO Item(
                           ItemId
                         , Name
                         , Created
                         , Modified
                         , ParentItemId
                         , ItemType
                         , Content
                         , ContentType
                         , SerialNumber
                         )
                      SELECT
                           @Identity
                         , @Name
                         , GETUTCDATE()
                         , GETUTCDATE()
                         , @Parent
                         , ItemType
                         , Content
                         , ContentType
                         , SerialNumber
                      FROM Item
                      WHERE ItemId = @ItemId";

                destID = Guid.NewGuid();
                Context.ExecuteNonQuery(
                    commandText,
                    "@Name", destName,
                    "@Parent", destFolder.ItemId,
                    "@ItemId", ItemId,
                    "@Identity", destID);

                destFolder.UpdateModified();

                if (this is IFolder)
                {
                    createdFolder = new DavFolder(
                        Context,
                        destID,
                        destFolder.ItemId,
                        destName,
                        destFolder.Path + EncodeUtil.EncodeUrlPart(destName) + "/",
                        DateTime.UtcNow,
                        DateTime.UtcNow);
                }
            }
            else 
            {
                // update existing destination
                destID = destItem.ItemId;

                string commandText = @"UPDATE Item SET
                                       Modified = GETUTCDATE()
                                       , ItemType = src.ItemType
                                       , ContentType = src.ContentType
                                       FROM (SELECT * FROM Item WHERE ItemId=@SrcID) src
                                       WHERE Item.ItemId=@DestID";

                Context.ExecuteNonQuery(
                    commandText,
                    "@SrcID", ItemId,
                    "@DestID", destID);

                // remove old properties from the destination
                Context.ExecuteNonQuery(
                    "DELETE FROM Property WHERE ItemID = @ItemID",
                    "@ItemID", destID);
            }

            // copy properties
            string command =
                @"INSERT INTO Property(ItemID, Name, Namespace, PropVal)
                  SELECT @DestID, Name, Namespace, PropVal
                  FROM Property
                  WHERE ItemID = @SrcID";

            Context.ExecuteNonQuery(
                command,
                "@SrcID", ItemId,
                "@DestID", destID);

            return createdFolder;
        }

        internal void MoveThisItem(DavFolder destFolder, string destName, DavFolder parent)
        {
            string command = 
                @"UPDATE Item SET
                     Name = @Name,
                     ParentItemId = @Parent
                  WHERE ItemId = @ItemID";

            Context.ExecuteNonQuery(
                command,
                "@ItemID", ItemId,
                "@Name", destName,
                "@Parent", destFolder.ItemId);

            parent.UpdateModified();
            destFolder.UpdateModified();
        }

        internal void DeleteThisItem()
        {
            DeleteThisItem(GetParent());
        }

        internal void DeleteThisItem(DavFolder parentFolder)
        {
            Context.ExecuteNonQuery(
                "DELETE FROM Item WHERE ItemId = @ItemID",
                "@ItemID", ItemId);

            if (parentFolder != null)
            {
                parentFolder.UpdateModified();
            }
        }

        internal void UpdateModified()
        {
            Context.ExecuteNonQuery(
                "UPDATE Item SET Modified = GETUTCDATE() WHERE ItemId = @ItemId",
                "@ItemId", ItemId);
        }

        protected string CurrentUserName
        {
            get { return Context.User != null ? Context.User.Identity.Name : string.Empty; }
        }
    }
}