using System;
using System.Configuration;
using System.IO;
using System.Net;
using System.Reflection;
using System.Security.Principal;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using System.Xml;
using ITHit.WebDAV.Server;

namespace WebDAVServer.Class1Listener
{
    partial class Service : ServiceBase
    {
        private Thread thread;

        private bool listening;

        private static readonly DigestAuthenticationProvider digestProvider = 
            new DigestAuthenticationProvider(getPasswordAndRoles);

        private static readonly string authType = 
            ConfigurationManager.AppSettings["AuthenticationType"];

        private readonly DavEngine engine;
        
        #region Public Members

        public Service(ILogger logger) : this(logger, false)
        {
            InitializeComponent();
        }

        public Service(ILogger logger, bool isApplication)
        {
            listening = isApplication;

            engine = new DavEngine
            {
                Logger = logger,
                OutputXmlFormatting = logger.IsDebugEnabled ? Formatting.Indented : Formatting.None,
            };
            var exePath = Path.GetDirectoryName(new Uri(Assembly.GetExecutingAssembly().CodeBase).LocalPath);
            engine.License = File.ReadAllText(Path.Combine(exePath, "License.lic"));
            // set custom handler to process GET requests to folders and display info about how to connect to server
            MyCustomGetHandler handler = new MyCustomGetHandler();
            handler.OriginalHandler = engine.RegisterMethodHandler("GET", handler);
        }

        public void ThreadProc()
        {
            string uriPrefix = ConfigurationManager.AppSettings["ListenerPrefix"];
            HttpListener listener = new HttpListener();
            listener.Prefixes.Add(uriPrefix);

            initAuthenticationScheme(listener);

            listener.Start();

            Console.WriteLine("Start listening on " + uriPrefix);
            Console.WriteLine("Press Control-C to stop listener...");

            while (listening)
            {
                IAsyncResult result = listener.BeginGetContext(ListenerCallback, listener);
                result.AsyncWaitHandle.WaitOne();
            }
        }

        public void ListenerCallback(IAsyncResult result)
        {
            HttpListener listener = (HttpListener)result.AsyncState;
            HttpListenerContext context = listener.EndGetContext(result);
            MacOsXPreprocessor.Process(context.Request); // fixes headers for Mac OS X v10.5.3 or later

            if (!userAutorized(context))
            {
                context.Response.StatusCode = 401;
                showLoginDialog(context, context.Response);
                closeResponse(context);
                return;
            }

            context.Response.SendChunked = false;
            using (var sqlDavContext = new DavContext(context, listener.Prefixes, context.User))
            {
               engine.Run(sqlDavContext);
            }

            if (context.Response.StatusCode == 401)
            {
                showLoginDialog(context, context.Response);
            }

            closeResponse(context);
        }

        private void closeResponse(HttpListenerContext context)
        {
            try
            {
                context.Response.Close();
            }
            catch
            {
                // client closed connection before the content was sent
            }
        }

        #endregion // Public Members

        #region ServiceBase Methods Override

        protected override void OnStart(string[] args)
        {
            listening = true;
            thread = new Thread(ThreadProc);
            thread.Start();
        }

        protected override void OnStop()
        {
            listening = false;
            thread.Abort();
            thread.Join();
        }

        #endregion // ServiceBase Methods Override

        #region Private Members

        private void initAuthenticationScheme(HttpListener listener)
        {
            switch (authType.ToLower())
            {
                case "none":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;
                    break;
                case "basic":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Basic;
                    break;
                case "digest":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;
                    break;
                case "ntlm":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Ntlm;
                    // Required to avoid buffering on a client side.
                    // Otherwise "This request requires buffering data to succeed" 
                    // exception may be thrown on a client side.
                    listener.UnsafeConnectionNtlmAuthentication = true;
                    break;
                case "kerberos":
                    listener.AuthenticationSchemes = AuthenticationSchemes.IntegratedWindowsAuthentication;
                    break;
            }
        }

        private static DigestAuthenticationProvider.PasswordAndRoles getPasswordAndRoles(
            string username)
        {
            if (username == "User1")
            {
                return new DigestAuthenticationProvider.PasswordAndRoles("pwd", new[] { "admin" });
            }

            return null;
        }

        private static void showLoginDialog(HttpListenerContext context, HttpListenerResponse response)
        {
            switch (authType.ToLower())
            {
                case "basic":
                    context.Response.AddHeader(
                        "WWW-Authenticate",
                        "Basic Realm=\"My WebDAV Server\"");
                    break;
                case "digest":
                    context.Response.AddHeader(
                        "WWW-Authenticate",
                         digestProvider.GetChallenge());
                    break;
                case "ntlm":
                    response.AddHeader("WWW-Authenticate", "NTLM");
                    break;
                case "kerberos":
                    response.AddHeader("WWW-Authenticate", "Negotiate");
                    break;
            }

            if (context.Request.HttpMethod.ToLower() != "options"
                && context.Request.HttpMethod.ToLower() != "head")
            {
                byte[] message = new UTF8Encoding().GetBytes("Access denied");
                context.Response.ContentLength64 = message.Length;
                context.Response.OutputStream.Write(message, 0, message.Length);
            }
        }

        private static bool userAutorized(HttpListenerContext context)
        {
            switch (authType.ToLower())
            {
                case "basic":
                    HttpListenerBasicIdentity identity = (HttpListenerBasicIdentity)context.User.Identity;
                    string username = identity.Name;

                    // workaround for Windows Vista Basic authentication.
                    // User name may be submitted in the following format: Machine\User.
                    int ind = username.LastIndexOf('\\');
                    if (ind > 0)
                    {
                        username = username.Remove(0, ind + 1);
                    }

                    return username == "User1" && identity.Password == "pwd";

                case "digest":
                    // workaround for Windows Server 2003 and XP.
                    // Digest Authorization header is not attched to OPTIONS requests.
                    if (context.Request.HttpMethod.ToLower() == "options")
                    {
                        return true;
                    }

                    return digestProvider.AuthenticateRequest(
                        context.Request.Headers,
                        context.Request.HttpMethod) != null;

                default:
                    return true;
            }
        }

        #endregion // Private Members
    }
}
