﻿//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (!Params.ACL)  return string.Empty; #>
//!<# if (Params.Storage != StorageType.Ntfs)  return string.Empty; #>
using System;
using System.Collections.Generic;
using System.DirectoryServices;
using System.DirectoryServices.AccountManagement;
using System.Linq;
using System.Security.Principal;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;
using IPrincipal = ITHit.WebDAV.Server.Acl.IPrincipal;

namespace ITHit.WebDav.Server.SampleModel.__NtfsStorage__.Acl
{
    /// <summary>
    /// Base class for users and groups.
    /// </summary>
    public abstract class PrincipalBase : IPrincipal
    {
        /// <summary>
        /// Encoded path to the parent folder.
        /// </summary>
        private readonly string parentPath;

        /// <summary>
        /// Initializes a new instance of the PrincipalBase class.
        /// </summary>
        /// <param name="principal">Instance of <see cref="Principal"/>.</param>
        /// <param name="parentPath">Encoded path to parent folder.</param>
        /// <param name="context">Instance of <see cref="DavContext"/>.</param>
        protected PrincipalBase(Principal principal, string parentPath, DavContext context)
        {
            this.Principal = principal;
            this.parentPath = parentPath;
            this.Context = context;
        }

        /// <summary>
        /// Gets corresponding <see cref="Principal"/>.
        /// </summary>
        public Principal Principal { get; private set; }

        /// <summary>
        /// Gets principal name.
        /// </summary>
        public string Name
        {
            get { return Principal.Name; }
        }

        /// <summary>
        /// Gets date when principal was created.
        /// </summary>
        public DateTime Created
        {
            get
            { 
                object o = ((DirectoryEntry)Principal.GetUnderlyingObject()).Properties["whenCreated"].Value;
                return o != null ? (DateTime)o : DateTime.MinValue;
            }
        }

        /// <summary>
        /// Gets date when principal was modified.
        /// </summary>
        public DateTime Modified
        {
            get
            {
                object o = ((DirectoryEntry)Principal.GetUnderlyingObject()).Properties["whenChanged"].Value;
                return o != null ? (DateTime)o : DateTime.MinValue;
            }
        }

        /// <summary>
        /// Gets principal's security identifier.
        /// </summary>
        public SecurityIdentifier Sid
        {
            get { return Principal.Sid; }
        }

        /// <summary>
        /// Gets encoded path to this principal.
        /// </summary>
        public abstract string Path { get; }

        /// <summary>
        /// Gets instance of <see cref="DavContext"/>.
        /// </summary>
        protected DavContext Context { get; private set; }

        /// <summary>
        /// Checks principal name for validity.
        /// </summary>
        /// <param name="name">Name to check.</param>
        /// <returns>Whether principal name is valid.</returns>
        public static bool IsValidUserName(string name)
        {
            char[] invChars = new[] { '"', '/', '\\', '[', ']', ':', ';', '|', '=', ',', '+', '*', '?', '<', '>' };
            return !invChars.Where(c => name.Contains(c)).Any();
        }

        //$<IPrincipal.GetGroupMembership
        /// <summary>
        /// Gets groups to which this principal belongs.
        /// </summary>
        /// <returns>Enumerable with groups.</returns>
        public IEnumerable<IPrincipal> GetGroupMembership()
        {
            return Principal.GetGroups().Select(group => (IPrincipal)new Group((GroupPrincipal)group, Context));
        }
        //$>

        /// <summary>
        /// Deletes the principal.
        /// </summary>
        /// <param name="multistatus">We don't use it currently as there are no child objects.</param>
        public void Delete(MultistatusException multistatus)
        {
            Context.PrincipalOperation(Principal.Delete);
        }

        /// <summary>
        /// Renames principal.
        /// </summary>
        /// <param name="destFolder">We don't use it as moving groups to different folder is not supported.</param>
        /// <param name="destName">New name.</param>
        /// <param name="multistatus">We don't use it as there're no child objects.</param>
        public void MoveTo(IItemCollection destFolder, string destName, MultistatusException multistatus)
        {
            if (destFolder.Path != parentPath)
            {
                throw new DavException("Moving principals is only allowed into the same folder", DavStatus.CONFLICT);
            }

            if (!IsValidUserName(destName))
            {
                throw new DavException("Principal name contains invalid characters", DavStatus.FORBIDDEN);
            }

            Context.PrincipalOperation(
                () => ((DirectoryEntry)Principal.GetUnderlyingObject()).Rename(destName));
        }

        public abstract void SetGroupMembers(IList<IPrincipal> members);

        public abstract IEnumerable<IPrincipal> GetGroupMembers();

        public abstract bool IsWellKnownPrincipal(WellKnownPrincipal wellknownPrincipal);

        public abstract IEnumerable<PropertyValue> GetProperties(IList<PropertyName> props, bool allprop);

        public abstract IEnumerable<PropertyName> GetPropertyNames();

        public abstract void UpdateProperties(
            IList<PropertyValue> setProps,
            IList<PropertyName> delProps,
            MultistatusException multistatus);

        public abstract void CopyTo(
            IItemCollection destFolder,
            string destName,
            bool deep,
            MultistatusException multistatus);
    }
}
