//!<#@ template language="C#" #>
//!<#@ samplevars Processor="DirectiveProcessor" #>
//!<# if (Params.Storage != StorageType.Ntfs)  return string.Empty; #>
using System;
using System.Configuration;
//!<# if (Params.ACL) { #>
using System.DirectoryServices.AccountManagement;
//!<# } #>
using System.IO;
using System.Linq;
using System.Net;
using System.Runtime.InteropServices;
using System.Security.Principal;
//!<# if (Params.CalDav) { #>
using System.Text.RegularExpressions;
//!<# } #>
using System.Web;

using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Acl;
using ITHit.WebDAV.Server.Quota;

//!<# if (Params.ACL) { #>
using ITHit.WebDav.Server.SampleModel.__NtfsStorage__.Acl;
//!<# } #>
//!<# if (Params.CalDav) { #>
// using ITHit.WebDav.Server.SampleModel.__NtfsStorage__.CalDav;
//!<# } #>

namespace ITHit.WebDav.Server.SampleModel.__NtfsStorage__
{
    /// <summary>
    /// Implementation of <see cref="DavContext"/>.
    /// Resolves hierarchy items by paths.
    /// </summary>
    public class DavContext : DavContextBase/*!<# if (Params.ACL) { #>!*/, IDisposable/*!<# } #>!*/
    {
        /// <summary>
        /// Path to folder where files are stored.
        /// </summary>
        private readonly string repositoryPath;

        /// <summary>
        /// Instance of <see cref="ILogger"/> to be used for logging.
        /// </summary>
        private readonly ILogger logger;

        /// <summary>
        /// Currently logged in identity.
        /// </summary>
        private readonly WindowsIdentity windowsIdentity;

        /// <summary>
        /// Disk full windows error code.
        /// </summary>
        private const int ERROR_DISK_FULL = 0x70;

        //!<# if (Params.ACL) { #>
        /// <summary>
        /// A <see cref="PrincipalContext"/> to be used for windows users operations.
        /// </summary>
        private PrincipalContext principalContext;
        //!<# } #>

        //!<# if (Params.Hosting == HostingEnvironment.IIS) { #>
        /// <summary>
        /// Initializes a new instance of the DavContext class.
        /// </summary>
        /// <param name="httpContext"><see cref="HttpContext"/> instance.</param>
        /// <param name="repositoryPath">Local path to repository.</param>
        /// <param name="logger"><see cref="ILogger"/> instance.</param>
        public DavContext(HttpContext httpContext, string repositoryPath, ILogger logger) : base(httpContext)
        {
            this.logger = logger;
            this.repositoryPath = repositoryPath;
            if (!Directory.Exists(repositoryPath))
            {
                logger.LogError("Repository path specified in Web.config is invalid.", null);
            }

            windowsIdentity = httpContext.User.Identity as WindowsIdentity;
        }
        //!<# } #>
      
        //!<# if (Params.Hosting == HostingEnvironment.Listener) { #>
        /// <summary>
        /// Initializes a new instance of the DavContext class.
        /// </summary>
        /// <param name="listenerContext"><see cref="HttpListenerContext"/> instance.</param>
        /// <param name="prefixes">Http listener prefixes.</param>
        /// <param name="repositoryPath">Local path to repository.</param>
        /// <param name="logger"><see cref="ILogger"/> instance.</param>
        public DavContext(
            HttpListenerContext listenerContext,
            HttpListenerPrefixCollection prefixes,
            string repositoryPath,
            ILogger logger)
            : base(listenerContext, prefixes)
        {
            this.logger = logger;
            this.repositoryPath = repositoryPath;
            if (!Directory.Exists(repositoryPath))
            {
                logger.LogError("Repository path specified in Web.config is invalid.", null);
            }

            windowsIdentity = listenerContext.User.Identity as WindowsIdentity;
        }
        //!<# } #>

        /// <summary>
        /// Gets repository path.
        /// </summary>
        public string RepositoryPath
        {
            get { return repositoryPath; }
        }

        /// <summary>
        /// Gets <see cref="ILogger"/> instance.
        /// </summary>
        public ILogger Logger
        {
            get { return logger; }
        }

        //!<# if (Params.ACL) { #>
        /// <summary>
        /// Gets domain of currently logged in user.
        /// </summary>
        public string Domain
        {
            get
            {
                int i = User.Name.IndexOf("\\");
                return i > 0 ? User.Name.Substring(0, i) : Environment.MachineName;
            }
        }
        //!<# } #>

        //!<# if (Params.ACL) { #>
        /// <summary>
        /// Gets currently authenticated user.
        /// </summary>
        public WindowsIdentity User
        {
            get
            {
                if (windowsIdentity != null && !windowsIdentity.IsAnonymous)
                {
                    return windowsIdentity;
                }

                if (AnonymousUser != null)
                {
                    return AnonymousUser;
                }

                throw new Exception("Anonymous user is not configured.");
            }
        }

        /// <summary>
        /// Gets or sets user configured as anonymous.
        /// </summary>
        public WindowsIdentity AnonymousUser { get; set; }

        /// <summary>
        /// Retrieves or creates <see cref="PrincipalContext"/> to be used for user related operations.
        /// </summary>
        /// <returns>Instance of <see cref="PrincipalContext"/>.</returns>
        public PrincipalContext GetPrincipalContext()
        {
            if (principalContext == null)
            {
                if (string.IsNullOrEmpty(Domain) ||
                    Environment.MachineName.Equals(Domain, StringComparison.InvariantCultureIgnoreCase))
                {
                    principalContext = new PrincipalContext(ContextType.Machine, Domain);
                }
                else
                {
                   principalContext = new PrincipalContext(ContextType.Domain);
                }
            }

            return principalContext;
        }

        /// <summary>
        /// Performs operation which creates, deletes or modifies user or group.
        /// Performs impersonification and exception handling.
        /// </summary>
        /// <param name="action">Action which performs action with a user or group.</param>
        public void PrincipalOperation(Action action)
        {
            using (impersonate())
            {
                try
                {
                    action();
                }
                catch (PrincipalOperationException ex)
                {
                    Logger.LogError("Principal operation failed", ex);
                    throw new DavException("Principal operation failed", ex);
                }
            }
        }

        /// <summary>
        /// Performs operation which queries, creates, deletes or modifies user or group.
        /// Performs impersonification and exception handling.
        /// </summary>
        /// <param name="func">Function to perform.</param>
        /// <typeparam name="T">Type of operation result.</typeparam>
        /// <returns>The value which <paramref name="func"/> returned.</returns>
        public T PrincipalOperation<T>(Func<T> func)
        {
            using (impersonate())
            {
                try
                {
                    return func();
                }
                catch (PrincipalOperationException ex)
                {
                    Logger.LogError("Principal operation failed", ex);
                    throw new DavException("Principal operation failed", ex);
                }
            }
        }
        //!<# } #>

        //$<DavContextBase.GetHierarchyItem
        /// <summary>
        /// Creates <see cref="IHierarchyItem"/> instance by path.
        /// </summary>
        /// <param name="path">Item relative path including query string.</param>
        /// <returns>Instance of corresponding <see cref="IHierarchyItem"/> or null if item is not found.</returns>
        public override IHierarchyItem GetHierarchyItem(string path)
        {
            path = path.Trim();
            //remove query string.
            int ind = path.IndexOf('?');
            if (ind > 0)
            {
                path = path.Remove(ind);
            }

            //Convert to local file system path by decoding every part, reversing slashes and appending
            //to repository root.
            string[] encodedParts = path.Split(new[] { "/" }, StringSplitOptions.RemoveEmptyEntries);
            string[] decodedParts = encodedParts.Select<string, string>(EncodeUtil.DecodeUrlPart).ToArray();
            string fullPath = Path.Combine(repositoryPath, string.Join("\\", decodedParts));

            path = path.Trim(new[] { '/' });

            //!<# if (Params.ACL) { #>
            //If this is /acl - return fake folder which contains users and groups.
            if (path == "acl")
            {
                return new AclFolder(this);
            }

            //if this is /acl/users - return fake folder which contains users.
            if (path == UserFolder.PREFIX)
            {
                return new UserFolder(this);
            }

            //if this is /acl/groups - return fake folder which contains groups.
            if (path == GroupFolder.PREFIX)
            {
                return new GroupFolder(this);
            }
            
            //if this is /acl/users/<user name> - return instance of User.
            if (path.StartsWith(UserFolder.PATH))
            {
                string name = EncodeUtil.DecodeUrlPart(path.Substring(UserFolder.PATH.Length));
                //we don't need an exception here - so check for validity.
                if (!PrincipalBase.IsValidUserName(name))
                {
                    return null;
                }

                return Acl.User.FromName(name, this);
            }

            //if this is /acl/groups/<group name> - return instance of Group.
            if (path.StartsWith(GroupFolder.PATH))
            {
                string name = EncodeUtil.DecodeUrlPart(path.Substring(GroupFolder.PATH.Length));
                if (!PrincipalBase.IsValidUserName(name))
                {
                    return null;
                }

                return Acl.Group.FromName(name, this);
            }
            //!<# } #>
            
            //!<# if (Params.CalDav) { #>
            /*
            if (path == "calendars")
            {
                return new CalendarsFolder(this);
            }

            if (path == "calendars/__uids")
            {
                return new UidsFolder(this);
            }

            if (path.StartsWith("calendars/users/"))
            {
                Match m = Regex.Match(path, "calendars/users/(?<user>[^/]+)/?(?<folder>[^/]+)?/?(?<file>[^/]+)?");
                if (m.Success)
                {
                    string userPath = null;
                    if (m.Groups["user"].Success)
                    {
                        userPath = Path.Combine(RepositoryPath, "calendars");
                        userPath = Path.Combine(userPath, m.Groups["user"].Value);
                        if (!Directory.Exists(userPath) && userExists(m.Groups["user"].Value))
                        {
                            Directory.CreateDirectory(Path.Combine(userPath, "calendar"));
                            Directory.CreateDirectory(Path.Combine(userPath, "inbox"));
                            Directory.CreateDirectory(Path.Combine(userPath, "outbox"));
                        }
                    }

                    if (m.Groups["file"].Success)
                    {
                        var localPath = Path.Combine(RepositoryPath, "calendars");
                        localPath = Path.Combine(localPath, m.Groups["user"].Value);
                        localPath = Path.Combine(localPath, m.Groups["folder"].Value);
                        localPath = Path.Combine(localPath, m.Groups["file"].Value);
                        
                        if (File.Exists(localPath))
                        {
                            return new CalendarResource(new FileInfo(localPath), this, path);
                        }
                    }
                    else if (m.Groups["folder"].Success)
                    {
                        var r = new CalFolder(this, m.Groups["user"].Value, m.Groups["folder"].Value);
                        if (r.Exists)
                        {
                            return r;
                        }
                    }
                    else
                    {
                        return new CalUserFolder(new DirectoryInfo(userPath), this, path);
                    }
                }
            }*/
            //!<# } #>
            
            //Many clients don't follow rule to specify folder path ending with slash,
            //and we cannot rely on it, so check if a folder is requested.
            var directory = new DirectoryInfo(fullPath);
            if (directory.Exists)
            {
                string correctFolderPath = string.IsNullOrEmpty(path) ? "/" : "/" + path + "/";
                return new DavFolder(directory, this, correctFolderPath);
            }
            
            //Check if a file is requested
            var file = new FileInfo(fullPath);
            if (file.Exists)
            {
                return new DavFile(file, this, "/" + path);
            }

            logger.LogDebug("Could not resolve path:" + fullPath);
            
            return null; // no hierarchy item that coresponds to path parameter was found in the repository
        }
        //$>

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine.
        /// </summary>
        /// <param name="item">Item on which operation is performed.</param>
        /// <param name="action">The action to be performed.</param>
        /// <param name="privilege">Privilege which is needed to perform the operation. If <see cref="UnauthorizedAccessException"/> is thrown
        /// this method will convert it to <see cref="NeedPrivilegesException"/> exception and specify this privilege in it.</param>
        public void FileOperation(IHierarchyItem item, Action action, Privilege privilege)
        {
            try
            {
                //!<# if (Params.ACL) { #>
                using (impersonate())
                {
                    action();
                }
                //!<# } #>
                //!<# if (!Params.ACL) { #>
                action();
                //!<# } #>
            }
            catch (UnauthorizedAccessException)
            {
                var ex = new NeedPrivilegesException("Not enough privileges");
                ex.AddRequiredPrivilege(item.Path, privilege);
                throw ex;
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine, except
        /// <see cref="UnauthorizedAccessException"/> which must be caught and translated manually.
        /// </summary>        
        /// <param name="action">The action to be performed.</param>
        public void FileOperation(Action action)
        {
            try
            {
                //!<# if (Params.ACL) { #>
                using (impersonate())
                {
                    action();
                }
                //!<# } #>
                //!<# if (!Params.ACL) { #>
                action();
                //!<# } #>
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        /// <summary>
        /// Performs file system operation with translating exceptions to those expected by WebDAV engine.
        /// </summary>
        /// <param name="item">Item on which operation is performed.</param>
        /// <param name="func">The action to be performed.</param>
        /// <param name="privilege">Privilege which is needed to perform the operation.
        /// If <see cref="UnauthorizedAccessException"/> is thrown  this method will convert it to
        /// <see cref="NeedPrivilegesException"/> exception and specify this privilege in it.</param>
        /// <typeparam name="T">Type of operation's result.</typeparam>
        /// <returns>Result returned by <paramref name="func"/>.</returns>
        public T FileOperation<T>(IHierarchyItem item, Func<T> func, Privilege privilege)
        {
            try
            {
                //!<# if (Params.ACL) { #>
                using (impersonate())
                {
                    return func();
                }
                //!<# } #>
                //!<# if (!Params.ACL) { #>
                return func();
                //!<# } #>
            }
            catch (UnauthorizedAccessException)
            {
                var ex = new NeedPrivilegesException("Not enough privileges");
                ex.AddRequiredPrivilege(item.Path, privilege);
                throw ex;
            }
            catch (IOException ex)
            {
                int hr = Marshal.GetHRForException(ex);
                if (hr == ERROR_DISK_FULL)
                {
                    throw new InsufficientStorageException();
                }

                throw new DavException(ex.Message, DavStatus.CONFLICT);
            }
        }

        //!<# if (Params.ACL) { #>
        /// <summary>
        /// Dispose everything we have.
        /// </summary>
        public void Dispose()
        {
            if (principalContext != null)
            {
                principalContext.Dispose();
            }
        }
        //!<# } #>

        //!<# if (Params.CalDav) { #>
        private bool userExists(string name)
        {
            return true;
        }
        //!<# } #>

        //!<# if (Params.ACL) { #>
        /// <summary>
        /// Impersonates current user.
        /// </summary>
        /// <returns>Impersonation context, which must be disposed to 'unimpersonate'</returns>
        private WindowsImpersonationContext impersonate()
        {
            return LogonUtil.DuplicateToken(User).Impersonate();
        }
        //!<# } #>
    }
}
